/****************************************************************

file swd_case1.c

Example Code searching for solutions of Sir Peter Swinnerton-Dyer's
Diophantine equation. An implementation of the hashing algorithm
developed by the authors.

Copyright (C) 2006 A.-S. Elsenhans and J. Jahnel

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Address of the authors

	A.-S. Elsenhans/ J. Jahnel
	Math. Institut der Universitaet
	Bunsenstrasze 3--5
	D-37073 Goettingen
        Germany

WWW     http://www.uni-math.gwdg.de/jahnel

*****************************************************************/


#include"common.h"

#define HASH_GRENZE  ((1U << HASH_BREITE) - 1)
#define PUFFER_ZAHL   (1U << PUFFER_BITS)
#define PUFFER_LEN    (1U << (HASH_BREITE - PUFFER_BITS))
#define LEN           (1U << HASH_BREITE)

/* The hash table. */
signed int                                 hash_c[LEN];

/* Buffers. There are PUFFER_ZAHL of them.
   The same buffers are used for reading and writing.
   Without "+173" a lot slower in out tests?!
   Other numbers instead of 173: Often almost as fast. */
unsigned long int  puffer[PUFFER_ZAHL][PUFFER_LEN+173];
/* fenster_ende[i] always points to the first unoccupied entry in
   the buffer puffer[i]. */
unsigned long int           *fenster_ende[PUFFER_ZAHL];

/* Table of the fourth powers modulo PAGE_PRIME1. */
unsigned int                     h4_mod_p[PAGE_PRIME1];

/* Table of the fourth roots modulo PAGE_PRIME1. */
int                        vierte_wurzeln[PAGE_PRIME1];

/* Counters. */
long int                                           anz;
long int                                        h_used;

/* Constants for the Chinese Remainder Theorem.
   Modules: PAGE_PRIME1, 1250. */
long int                          eins_null, null_eins;

/* The number of clearings. Separate counters for each buffer.
   Necessary for the post processing. */
int                      puffer_leerungen[PUFFER_ZAHL];


/* The hash function. */
#define HASH(a)        (((unsigned int) (a)) >> (32 - HASH_BREITE))
/* Part of the hash function used for presorting. */
#define PUFFER_NR(a)   (((unsigned int) (a)) >> (32 - PUFFER_BITS))

/* CONTROL_FILTER = 2**n - 1 where n <= 31.
   For n = 32, the control value could become negative and -1 would
   lose the extra meaning "Place unoccupied". */
#define CONTROL_FILTER ((1U << 31) - 1)
/* The control function. */
#define CONTROL(a)     ((signed int) (((a) >> 32) & CONTROL_FILTER))


/* z = (a mod PAGE_PRIME1) and z = (b mod 1250). */
#define CHINESE_1250(a, b)                                        \
((((long int) a) * eins_null +                                    \
  ((long int) b) * null_eins) % (1250*PAGE_PRIME1))


/* Tests for PAGE_PRIME1.
   Tabulation of the fourth powers modulo PAGE_PRIME1.
   Initialization of the constants for the Chinese Remainder
   Theorem. */
void init () {
 long int  i, t;

 /* Testing PAGE_PRIME1. */
 if (1250*PAGE_PRIME1 <= SUCHWEITE) {
  printf ("PAGE_PRIME1 must be bigger than SUCHWEITE/1250.");
  exit (0);
 }
 if ((PAGE_PRIME1 % 4) != 3) {
  printf ("PAGE_PRIME1 must be 3 modulo 4.\n");
  exit (0);
 }
 for (i = 3; i*i <= PAGE_PRIME1; i += 2) {
  if (PAGE_PRIME1 % i == 0) {
   printf ("PAGE_PRIME1 must a prime number.\n");
   exit (0);
  }
 }

 /* Initialization of the counter for coincidences. */
 anz = 0;

 /* Tabulate the fourth powers modulo PAGE_PRIME1. */
 for (i = 0; i < PAGE_PRIME1; i++) {
  t = (i*i) % PAGE_PRIME1;
  t = (t*t) % PAGE_PRIME1;
  h4_mod_p[i] = t;           /* Fourth powers modulo PAGE_PRIME1. */
 }

 for (i = 0; i < PAGE_PRIME1; i++)
  vierte_wurzeln[i] = -1;

 for (i = 0; i < PAGE_PRIME1; i++)
  vierte_wurzeln[h4_mod_p[i]] = i;

 /* Initialize the constants for the Chinese Remainder Theorem. */
 eins_null = 1;
 while ((eins_null % 1250) != 0)
  eins_null += PAGE_PRIME1;
 null_eins = 1;
 while ((null_eins % PAGE_PRIME1) != 0)
  null_eins += 1250;
}


/* CLEANING UP */

/* Declare the buffers to be empty.
   Initialization of puffer_leerungen for a new residue class. */
void puffer_init () {
 int  i;

 for (i = 0; i < PUFFER_ZAHL; i++) {
  fenster_ende[i] = &puffer[i][0];
  puffer_leerungen[i] = 0;
 }
}


/* Declare the hash table to be empty. */
void hashtabelle_init () {
 int  i;

 h_used = 0;

 for (i = 0; i < LEN; i++)
  hash_c[i] = -1;
}


/* WRITING */

/* Inclusion of a single pair into the buffer. */
inline void z_und_x_schreiben (int z, int x) {
 unsigned long int  z_lang, x_lang, rechte_seite;
 unsigned int                                  t;

 if ((z <= SUCHWEITE) &&
     (((x+z) & 7) == 0 || ((x-z) & 7) == 0) &&
     ((x % 3 == 0) || (z % 3 == 0) ||
      ((x+z) % 81 == 0) || ((x-z) % 81 == 0)) &&
     (x != z)) {

  z_lang = z;
  x_lang = x;

  rechte_seite =
   z_lang*z_lang*z_lang*z_lang - x_lang*x_lang*x_lang*x_lang;
  /* modulo 2**64. */

  t = PUFFER_NR (rechte_seite);

  /* This means overflow during the writing step.
     Should never happen. */
  if (fenster_ende[t] >= &puffer[t][0] + PUFFER_LEN) {
   printf ("Overflow in one of the buffers!\n");
   exit (0);
  }

  /* Write rechte_seite into the buffer. */
  *(fenster_ende[t]) = rechte_seite;
  fenster_ende[t]++;

  h_used++;
 }
}


/* Build-up of the buffer for writing.
   We assume that 1250*PAGE_PRIME1 > SUCHWEITE. */
void schreibdaten_erzeugen_und_puffern (int rkl) {
 int       z0, z1, x, x0;
 int  vw1, vw2, vw3, vw4;

 for (x0 = 1; x0 < 1250; x0 += 2) {
  if ((x0 % 5) != 0) {
   vw1 = x0 % 1250;
   vw2 = 1250 - vw1;
   vw3 = (625 + (182 * x0)) % 1250;
   vw4 = 1250 - vw3;
   /* We have z**4 = 1 mod 625 if and only if z = \pm 1
      or z = \pm182. We add 625 since 182 * x0 is even and z belongs
      to an odd residue class modulo 1250. */

   for (x = x0; x <= SUCHWEITE; x += 1250) {
    z0 = (rkl + h4_mod_p[x % PAGE_PRIME1]) % PAGE_PRIME1;
    z0 = vierte_wurzeln[z0];

    /* If there exists a fourth root modulo PAGE_PRIME1. */
    if (z0 != -1) {
     z1 = CHINESE_1250 (z0, vw1);
     z_und_x_schreiben (z1, x);
     if (z1 != 0) {
      z1 = 1250*PAGE_PRIME1 - z1;
      z_und_x_schreiben (z1, x);
     }

     z1 = CHINESE_1250 (z0, vw2);
     z_und_x_schreiben (z1, x);
     if (z1 != 0) {
      z1 = 1250*PAGE_PRIME1 - z1;
      z_und_x_schreiben (z1, x);
     }

     z1 = CHINESE_1250 (z0, vw3);
     z_und_x_schreiben (z1, x);
     if (z1 != 0) {
      z1 = 1250*PAGE_PRIME1 - z1;
      z_und_x_schreiben (z1, x);
     }

     z1 = CHINESE_1250 (z0, vw4);
     z_und_x_schreiben (z1, x);
     if (z1 != 0) {
      z1 = 1250*PAGE_PRIME1 - z1;
      z_und_x_schreiben (z1, x);
     }
    } /* End-if to if (z0 != -1). */
   }  /* To for (x = x0; ... ). */
  }   /* End-if to if ((x0 % 5) != 0). */
 }    /* To for (x0 = 1; ... ). */
}


/* Creation of the hash table from the buffer. */
void hashtabelle_schreiben () {
 int                                i, h;
 unsigned long int  rechte_seite, *index;

 for (i = 0; i < PUFFER_ZAHL; i++) {
  for (index = &puffer[i][0]; index < fenster_ende[i]; index++) {
   rechte_seite = *index;         /* rechte_seite = puffer[i][k] */
   h = HASH (rechte_seite);

   /* Resolution of collisions. */
   while (hash_c[h] != -1)
    h = ((h + 1) & HASH_GRENZE);   /* (h+1) mod (2**HASH_BREITE) */

   /* Inclusion. */
   hash_c[h] = CONTROL (rechte_seite);
  }
 }
 /* Hash built. */
}


/* READING */

/* One step towards building up the buffers for reading.
   y is fixed, w is free in one residue class.
   Start the actual reading when a buffer fills up. */
void restklasse_von_w_mit_y_fest_testen (int w, int y, int klasse) {
 unsigned long int  linke_seite, w_lang, y_lang, y4;
 unsigned int                                  w_10;
 signed int                                       f;

 w_10 = w;
 while ((w_10 % 10) != 0)
  w_10 += PAGE_PRIME1;

 y_lang = y;
 y4 = 2 * (y_lang*y_lang) * (y_lang*y_lang);     /* modulo 2**64. */

 for (w_lang = w_10; w_lang <= SUCHWEITE; w_lang += 10*PAGE_PRIME1)
 {
  linke_seite =  y4 - 4 * (w_lang*w_lang) * (w_lang*w_lang);
  /* modulo 2**64. */

  f = PUFFER_NR (linke_seite);

  /* If the buffer is full. */
  if (fenster_ende[f] >= &puffer[f][0] + PUFFER_LEN) {
   einen_puffer_abarbeiten (f, klasse);
   /* This leaves us with an empty buffer. */
  }

  *(fenster_ende[f]) = linke_seite;
  fenster_ende[f]++;
 } /* End of the for-loop for w. */
}


/* Build-up of the buffers for reading.
   The data in the buffers are actually read automatically every
   time when a buffer fills up. */
void lesedaten_erzeugen_und_lesen (int klasse) {
 int  y0, y, w0, t;

 for (w0 = 0; w0 < PAGE_PRIME1; w0++) {
  t = (klasse + 4 * h4_mod_p[w0 % PAGE_PRIME1]) % PAGE_PRIME1;
  if (t % 2 == 0)
   t = t / 2;
  else
   t = (t + PAGE_PRIME1) / 2;

  y0 = vierte_wurzeln[t];

  if (y0 != - 1) { /* If there exists a fourth root of t. */
   y = y0;
   while ((y % 10) != 0)
    y += PAGE_PRIME1;
   while (y <= SUCHWEITE) {
    restklasse_von_w_mit_y_fest_testen (w0, y, klasse);
    y += 10*PAGE_PRIME1;
   }

   if (y0 > 0) { /* If the fourth root of t is different from 0.
                    In this case, there are exactly two of them. */
    y = PAGE_PRIME1 - y0;
    while ((y % 10) != 0)
     y += PAGE_PRIME1;
    while (y <= SUCHWEITE) {
     restklasse_von_w_mit_y_fest_testen (w0, y, klasse);
     y += 10*PAGE_PRIME1;
    }
   }  /* End-if to if (y0 > 0). */
  }   /* End-if to if (y0 != -1). */
 }    /* To for (w0 = 0; ... ). */
}


/* All data in buffer number f_nr are looked up in the hash table.
   When a coincidence is detected, post processing is called.

   Afterwards, the buffer is declared to be empty.

   This function is called every time for a single buffer when it
   is full. At the very end, i.e. when the last reading data have
   been written into the buffer, it is called for all buffers. */
void einen_puffer_abarbeiten (int f_nr, int klasse) {
 unsigned long int  linke_seite, *index;
 signed int                           c;
 int                                  t;

 for (index = &puffer[f_nr][0]; index < fenster_ende[f_nr]; index++)
 {
  linke_seite = *index;          /* linke_seite = puffer[f_nr][k] */
  t = HASH (linke_seite);

  /* The actual reading. */
  /* Optimization for speed. Some sort of loop unrolling.
     If we hit an unoccupied place then that reading access finishes
     directly at this line. */
  if (hash_c[t] >= 0) {
   c = CONTROL (linke_seite);
   do {
    /* Found an interesting value. */
    if (hash_c[t] == c)
     post_proc (linke_seite, klasse, f_nr,
       PUFFER_LEN * puffer_leerungen[f_nr]
      + (index - &puffer[f_nr][0]));
     /* The last operand is the absolute buffer position which would
        have been reached without any clearings. */
    t = ((t + 1) & HASH_GRENZE);    /* (t+1) mod (2**HASH_BREITE) */
   } while (hash_c[t] != -1);     /* Observe possible collisions. */
  }  /* End-if for if (hash_c[t] >= 0). */
 }   /* To for (index = &puffer[f_nr][0]; ... ). */

 /* Declare the buffer to be empty. */
 fenster_ende[f_nr] = &puffer[f_nr][0];
 puffer_leerungen[f_nr]++;  /* Count the number of all clearings. */
}


/* Looking up all buffered data in the hash table.
   Clearing all buffers. */
void alle_puffer_abarbeiten (int klasse) {
 int  i;

 for (i = 0; i < PUFFER_ZAHL; i++)
  if (fenster_ende[i] > &puffer[i][0])
   einen_puffer_abarbeiten (i, klasse);
}


/* The outer loop. */
int main (int argc, char *argv[]) {
 int  rkl, start, ende;

 if (argc != 3) {
  printf ("Call needs 2 parameters.\n");
  exit (0);
 }
 start = atoi (argv[1]);
 ende = atoi (argv[2]);
 if (start < 0)
  start = 0;
 if (ende > (PAGE_PRIME1 - 1))
  ende = PAGE_PRIME1 - 1;
 if (start > ende) {
  printf ("Range from %d to %d is empty.\n", start, ende);
  exit (0);
 }

 printf ("Range from %d to %d.\n", start, ende);
 printf ("Initialize tables of powers and roots.\n");
 init ();
 printf ("Initialization finished.\n");

 for (rkl = start; rkl <= ende; rkl++) {
  /* Clearing up. */
  hashtabelle_init ();
  puffer_init ();

  /* Writing. */
  /* Build up the buffer in residue class rkl. */
  schreibdaten_erzeugen_und_puffern (rkl);
  /* Creating the hash table. */
  hashtabelle_schreiben ();
  printf ("Searching in residue class %d. ", rkl);
  printf ("%ld entries in the hash table.\n", h_used);

  /* Clearing up. */
  puffer_init ();

  /* Reading. */
  /* Build up the buffers in residue class rkl.
     The actual reading is done automatically in blocks
     corresponding to full buffers. */
  lesedaten_erzeugen_und_lesen (rkl);
  /* Clear all the buffers and read. */
  alle_puffer_abarbeiten (rkl);
  fflush (stdout);
 }

 printf ("Totally %ld coincidences detected.\n", anz);
 return 0;
}


/* POST PROCESSING */

/* The whole code below is only for post processing.
   It is not called too often. */

/* This function is one step of the simulated build-up of the
   buffers for reading. y is fixed, w is free in one residue
   class. */
int restklasse_von_w_mit_y_fest_probew_testen
   (int w, int y, int klasse, int f_nr, int i, int *puffer_index,
    int *w_gefunden, int *y_gefunden) {
 unsigned long int  linke_seite, w_lang, y_lang, y4;
 unsigned int                                  w_10;
 signed int                                       f;

 w_10 = w;
 while ((w_10 % 10) != 0)
  w_10 += PAGE_PRIME1;

 y_lang = y;
 y4 = 2 * (y_lang*y_lang) * (y_lang*y_lang);     /* modulo 2**64. */

 for (w_lang = w_10; w_lang <= SUCHWEITE; w_lang += 10*PAGE_PRIME1)
 {
  linke_seite = y4 - 4 * (w_lang*w_lang) * (w_lang*w_lang);
  /* modulo 2**64. */
  f = PUFFER_NR (linke_seite);
  if (f == f_nr) {
   (*puffer_index)++;
   if (*puffer_index == i) {
    *w_gefunden = w_lang;
    *y_gefunden = y;
    return 0;
   }
  }
 } /* End of the for-loop for w. */

 /* Return value 1 means "Not yet found".
    We return 0 as soon as the correct absolute buffer position is
    reached again. This means, immediately when x and y are
    found. */
 return 1;
}


/* Simulation of the build-up of the buffer for reading.

   This function is similar to lesedaten_erzeugen_und_lesen.
   It does, however, never write into memory. It only counts how
   often an entry would be written into buffer number f_nr.
   It finds the pair (w, y) which generates the i-th entry into
   buffer number f_nr. */
int lesepufferaufbau_simulieren (int klasse, int f_nr, int i,
    int *w_gefunden, int *y_gefunden) {
 int  y0, y, w0, t, puffer_index;

 puffer_index = -1;

 for (w0 = 0; w0 < PAGE_PRIME1; w0++) {
  t = (klasse + 4 * h4_mod_p[w0 % PAGE_PRIME1]) % PAGE_PRIME1;
  if (t % 2 == 0)
   t = t / 2;
  else
   t = (t + PAGE_PRIME1) / 2;

  y0 = vierte_wurzeln[t];

  if (y0 != - 1) { /* If there exists a fourth root of t. */
   y = y0;
   while ((y % 10) != 0)
    y += PAGE_PRIME1;
   while (y <= SUCHWEITE) {
    /* Return value 0 means "found". */
    if (restklasse_von_w_mit_y_fest_probew_testen
        (w0, y, klasse, f_nr, i, &puffer_index,
         w_gefunden, y_gefunden) == 0)
     return 0;
    y += 10*PAGE_PRIME1;
   }

   if (y0 > 0) { /* If the fourth root of t is different from 0.
                    In this case, there are exactly two of them. */
    y = PAGE_PRIME1 - y0;
    while ((y % 10) != 0)
     y += PAGE_PRIME1;
    while (y <= SUCHWEITE) {
     /* Return value 0 means "found". */
     if (restklasse_von_w_mit_y_fest_probew_testen
         (w0, y, klasse, f_nr, i, &puffer_index,
          w_gefunden, y_gefunden) == 0)
      return 0;
     y += 10*PAGE_PRIME1;
    }
   }  /* End-if to if (y0 > 0). */
  }   /* End-if to if (y0 != -1). */
 }    /* To for (w0; ... ). */

 /* This line should never be reached. When (w, y) is reconstructed
    then return 0 is carried out immediately. The return value 1
    allows easily to detect an error. */
 return 1;
}


/* This function calculates for a pair (z, x) whether the control
   value is equal to c. If this happens then an output is
   generated. */
inline void z_und_x_testen (int x, int y, int z, int w, signed int c) {
 unsigned long int x_lang, y_lang, z_lang, w_lang;
 unsigned long int           argh, argl, lsh, lsl;
 signed int                                c_wert;

 x_lang = x;
 y_lang = y;
 z_lang = z;
 w_lang = w;

 if (z <= SUCHWEITE) {
  umul_ppmm (lsh, lsl, z_lang * z_lang, z_lang * z_lang);
  umul_ppmm (argh, argl, x_lang * x_lang, x_lang * x_lang);
  sub_ddmmss (lsh, lsl, lsh, lsl, argh, argl);
  /* (lsh, lsl) = z**4 - x**4. */

  c_wert = CONTROL (lsl);

  if (c == c_wert) {
   printf ("x = %d, y = %d, z = %d, w = %d.\n", x, y, z, w);

   umul_ppmm (argh, argl, (w_lang * w_lang) << 1,
                          (w_lang * w_lang) << 1);
   add_ssaaaa (lsh, lsl, lsh, lsl, argh, argl);
   umul_ppmm (argh, argl, y_lang * y_lang, y_lang * y_lang);
   add_ssaaaa (argh, argl, argh, argl, argh, argl);
   sub_ddmmss (lsh, lsl, lsh, lsl, argh, argl);
   /* (lsh, lsl) = z**4 + 4 w**4 - x**4 - 2 y**4. */

   add_ssaaaa (lsh, lsl, lsh, lsl, lsh, lsl);
   /* This does not cause an overflow as long as
                SUCHWEITE < [2**128/10]**(1/4) = 2.41523760e9. */
   printf ("The hash deviation is %ld.\n", ((long int) lsl) / 2);
   printf ("The deviation in size is %ld * 2**63.\n", lsh);
   anz++;
  }
 }
}


/* Search for z and x such that the control value fits with ls. */
long int z_und_x_suchen (unsigned long int ls, int klasse,
         int w_gefunden, int y_gefunden) {
 long int              anz_alt;
 int             z0, z1, x, x0;
 int        vw1, vw2, vw3, vw4;
 signed int                  c;

 anz_alt = anz;
 c = CONTROL (ls);

 for (x0 = 1; x0 < 1250; x0 += 2) {
  if ((x0 % 5) != 0) {
   vw1 = x0 % 1250;
   vw2 = 1250 - vw1;
   vw3 = (625 + (182 * x0)) % 1250;
   vw4 = 1250 - vw3;

   for (x = x0; x <= SUCHWEITE; x += 1250) {
    z0 = (klasse + h4_mod_p[x % PAGE_PRIME1]) % PAGE_PRIME1;
    z0 = vierte_wurzeln[z0];

    /* If there exists a fourth root modulo PAGE_PRIME1. */
    if (z0 != -1) {
     z1 = CHINESE_1250 (z0, vw1);
     z_und_x_testen (x, y_gefunden, z1, w_gefunden, c);
     if (z1 != 0) {
      z1 = 1250*PAGE_PRIME1 - z1;
      z_und_x_testen (x, y_gefunden, z1, w_gefunden, c);
     }

     z1 = CHINESE_1250 (z0, vw2);
     z_und_x_testen (x, y_gefunden, z1, w_gefunden, c);
     if (z1 != 0) {
      z1 = 1250*PAGE_PRIME1 - z1;
      z_und_x_testen (x, y_gefunden, z1, w_gefunden, c);
     }

     z1 = CHINESE_1250 (z0, vw3);
     z_und_x_testen (x, y_gefunden, z1, w_gefunden, c);
     if (z1 != 0) {
      z1 = 1250*PAGE_PRIME1 - z1;
      z_und_x_testen (x, y_gefunden, z1, w_gefunden, c);
     }

     z1 = CHINESE_1250 (z0, vw4);
     z_und_x_testen (x, y_gefunden, z1, w_gefunden, c);
     if (z1 != 0) {
      z1 = 1250*PAGE_PRIME1 - z1;
      z_und_x_testen (x, y_gefunden, z1, w_gefunden, c);
     }
    } /* End-if to if (z0 != -1). */
   }  /* To for (x = x0; ... ). */
  }   /* End-if to if ((x0 % 5) != 0). */
 }    /* To for (x0 = 0; ... ). */

 /* The return value is the number of the pairs (z, x) fitting. */
 return (anz - anz_alt);
}


/* Post processing.

   Expects linke_seite in ls, (ls mod PAGE_PRIME1) in klasse, the
   number of the window in f_nr, and the absolute position of the
   entry causing the coincidence in i. */
void post_proc (unsigned long int ls, int klasse, int f_nr, int i) {
 int  w_gef, y_gef;

 /* Coincidence in residue class klasse, linke_seite = ls. */

 /* Reconstruction of w and y.
    Exiting with an error message if (w, y) cannot be
    reconstructed. */
 if (lesepufferaufbau_simulieren (klasse, f_nr, i, &w_gef, &y_gef))
 {
  printf ("Error in the buffer for reading. Exiting.\n");
  exit (0);
 }

 /* Reconstruction of z und x. It is possible that more than one
    pair is found. */
 if (z_und_x_suchen (ls, klasse, w_gef, y_gef) == 0) {
  /* Exiting with an error message if no pair (z, x) is found. */
  printf ("Error! Exiting.\n");
  exit (0);
 }
}
